@@ -2,9 +2,14 @@ import inspect
 import traceback
 from pathlib import Path
 
+import dash_ace_persistent
 import dash_core_components as dcc
+import dash_dangerously_set_inner_html
 import dash_html_components as html
 import networkx as nx
+import numpy as np
+import pandas as pd
+import plotly.express as px
 from dash import Dash
 from dash.dependencies import Input, Output, State
 from dash_interactive_graphviz import DashInteractiveGraphviz
@@ -12,23 +17,23 @@ from dash_split_pane import DashSplitPane
 from dash_treebeard import DashTreebeard
 from fn_graph.calculation import (
     NodeInstruction,
-    get_execution_instructions,
     calculate_collect_exceptions,
+    get_execution_instructions,
 )
 from fn_graph.profiler import Profiler
-
-__package__ = "fn_graph_studio"
+from pygments import highlight
+from pygments.formatters import HtmlFormatter
+from pygments.lexers import PythonLexer
 
 from .parameter_editor import (
-    parameter_widgets,
-    get_variable_parameter_keys,
     get_variable_parameter_ids,
+    get_variable_parameter_keys,
+    parameter_widgets,
 )
 from .result_renderers import add_default_renderers
 
-import pandas as pd
-import numpy as np
-import plotly.express as px
+__package__ = "fn_graph_studio"
+
 
 # Load up embedded styles
 # We embed the styles directly in the template ofr portabilities sake
@@ -36,6 +41,9 @@ import plotly.express as px
 with open(Path(__file__).parent / "styles.css") as f:
     styles = f.read()
 
+with open(Path(__file__).parent / "highlight.css") as f:
+    highlight_styles = f.read()
+
 
 def BasePane(default_style):
     def wrapper(*args, style=None, **kwargs):
@@ -88,7 +96,7 @@ VStack = BasePane(
 class BaseStudio:
     def __init__(self, app, composer, renderers=None):
 
-        app.title = "Fn Compose Studio"
+        app.title = "Fn Graph Studio"
 
         app.layout = self.layout(composer)
 
@@ -104,6 +112,8 @@ class BaseStudio:
                 <style>
                 """
             + styles
+            + "\n\n"
+            + highlight_styles
             + """
                 </style>
             </head>
@@ -157,6 +167,7 @@ class BaseStudio:
         @app.callback(
             Output(component_id="graphviz-viewer", component_property="dot_source"),
             [
+                Input(component_id="node-name-filter", component_property="value"),
                 Input(component_id="graph-display-options", component_property="value"),
                 Input(component_id="graph-neighbourhood", component_property="value"),
                 Input(
@@ -170,6 +181,7 @@ class BaseStudio:
             ],
         )
         def populate_graph_with_composer(
+            node_name_filter,
             graph_display_options,
             graph_neighbourhood,
             graph_neighbourhood_size,
@@ -178,6 +190,7 @@ class BaseStudio:
         ):
             return self.populate_graph(
                 composer,
+                node_name_filter,
                 graph_display_options,
                 graph_neighbourhood,
                 graph_neighbourhood_size,
@@ -186,6 +199,33 @@ class BaseStudio:
                 parameter_values,
             )
 
+        @app.callback(
+            Output(component_id="function-tree", component_property="data"),
+            [Input(component_id="node-name-filter", component_property="value")],
+        )
+        def populate_tree_with_composer(node_name_filter):
+            if node_name_filter:
+                matching_nodes = [
+                    key
+                    for key in composer.functions().keys()
+                    if node_name_filter.strip().lower() in key.lower()
+                ]
+                tree = composer.subgraph(matching_nodes)._build_name_tree()
+            else:
+                tree = composer._build_name_tree()
+
+            def format_tree(key, value):
+                if isinstance(value, str):
+                    return {"name": key, "key": value}
+                else:
+                    return {
+                        "name": key,
+                        "key": key,
+                        "children": [format_tree(k, v) for k, v in value.items()],
+                    }
+
+            return format_tree("_root_", tree)
+
         sidebar_components = self.sidebar_components(composer)
 
         @app.callback(
@@ -247,79 +287,72 @@ class BaseStudio:
         )
 
     def function_tree(self, component_id, composer):
-        tree = composer._build_name_tree()
-
-        def format_tree(key, value):
-            if isinstance(value, str):
-                return {"name": key, "key": value}
-            else:
-                return {
-                    "name": key,
-                    "key": key,
-                    "children": [format_tree(k, v) for k, v in value.items()],
-                }
-
         return Scroll(
-            DashTreebeard(id=component_id, data=format_tree("_root_", tree), persistence=True)
+            DashTreebeard(
+                id=component_id,
+                data={"name": "_root_", "key": "_root_", "children": []},
+                persistence="tree",
+            )
         )
 
     def function_graph(self):
+        label_width = 50
+
+        def Label(text):
+            return html.Strong(
+                f"{text}: ", style=dict(display="inline-block", width=label_width)
+            )
+
         return VStack(
             [
-                html.Div(
-                    [
-                        html.Strong("Filter: "),
-                        dcc.Checklist(
-                            id="graph-neighbourhood",
-                            options=[
-                                {"label": "All", "value": "all"},
-                                {"label": "Ancestors", "value": "ancestors"},
-                                {"label": "Descendants", "value": "descendants"},
-                                {"label": "Neighbours", "value": "neighbours"},
-                            ],
-                            value=["all"],
-                            persistence=True,
-                            labelStyle=dict(paddingLeft=10),
-                            style=dict(display="inline"),
-                        ),
-                    ]
+                dcc.Checklist(
+                    id="graph-neighbourhood",
+                    options=[
+                        {"label": "All", "value": "all"},
+                        {"label": "Ancestors", "value": "ancestors"},
+                        {"label": "Descendants", "value": "descendants"},
+                        {"label": "Neighbours", "value": "neighbours"},
+                    ],
+                    value=["all"],
+                    persistence=True,
+                    labelStyle=dict(paddingRight=10),
+                    inputStyle=dict(marginRight=2),
+                    style=dict(display="inline", marginBottom="5px"),
                 ),
-                html.Div(
-                    [
-                        html.Strong("Neighbourhood size: "),
-                        dcc.Input(
-                            id="graph-neighbourhood-size",
-                            type="number",
-                            value=1,
-                            min=0,
-                            persistence=1,
-                            style=dict(border="1px solid lightgrey"),
-                        ),
-                    ]
+                dcc.Input(
+                    id="graph-neighbourhood-size",
+                    type="number",
+                    value=1,
+                    min=0,
+                    persistence=1,
+                    style=dict(
+                        border="1px solid lightgrey", width="100%", marginBottom="5px"
+                    ),
                 ),
-                html.Div(
-                    [
-                        html.Strong("Display: "),
-                        dcc.Checklist(
-                            id="graph-display-options",
-                            options=[
-                                {"label": "Flatten", "value": "flatten"},
-                                {"label": "Parameters", "value": "parameters"},
-                                {"label": "Links", "value": "links"},
-                                {"label": "Caching", "value": "caching"},
-                            ],
-                            persistence=1,
-                            labelStyle=dict(paddingLeft=10),
-                            style=dict(display="inline"),
-                        ),
-                    ]
+                dcc.Checklist(
+                    id="graph-display-options",
+                    options=[
+                        {"label": "Flatten", "value": "flatten"},
+                        {"label": "Parameters", "value": "parameters"},
+                        {"label": "Links", "value": "links"},
+                        {"label": "Caching", "value": "caching"},
+                    ],
+                    persistence=1,
+                    labelStyle=dict(paddingRight=10),
+                    inputStyle=dict(marginRight=2),
                 ),
                 Pane(
                     DashInteractiveGraphviz(id="graphviz-viewer", persistence=False),
                     style=dict(flexGrow=1),
                 ),
             ],
-            style=dict(height="100%", width="100%", position="absolute", padding="5px"),
+            style=dict(
+                height="100%",
+                width="100%",
+                position="absolute",
+                padding="5px",
+                paddingTop="0px",
+            ),
         )
 
     def parameters(self, composer):
@@ -345,12 +378,19 @@ class BaseStudio:
         }
 
     def result_processor(self):
-        return dcc.Textarea(
+        return dash_ace_persistent.DashAceEditor(
             id="result-processor",
+            value="",
+            theme="github",
+            mode="python",
+            tabSize=2,
             placeholder="e.g. result.query(....)",
-            rows=5,
-            style=dict(width="100%", border="none"),
-            persistence=1,
+            maxLines=10,
+            minLines=5,
+            showGutter=False,
+            persistence=True,
+            highlightActiveLine=False,
+            debounceChangePeriod=1000,
         )
 
     def sidebar(self, composer):
@@ -363,18 +403,34 @@ class BaseStudio:
             [
                 HStack(
                     [
-                        html.Strong("Navigator: "),
-                        dcc.RadioItems(
+                        dcc.Tabs(
                             id="explorer-selector",
-                            options=explorer_options,
-                            labelStyle=dict(paddingLeft=10),
-                            value="",
+                            value="tab-1-example",
+                            parent_className="custom-tabs",
+                            className="custom-tabs-container",
+                            children=[
+                                dcc.Tab(label=option["label"], value=option["value"])
+                                for option in explorer_options
+                            ],
                             persistence=True,
-                        ),
+                        )
                     ],
-                    style=dict(
-                        flexShrink=0, padding="5px", borderBottom="1px solid lightgrey"
-                    ),
+                    style=dict(flexShrink=0),
+                ),
+                Pane(
+                    dcc.Input(
+                        id="node-name-filter",
+                        type="text",
+                        placeholder="Filter nodes",
+                        value="",
+                        persistence=True,
+                        style=dict(
+                            border="1px solid lightgrey",
+                            width=" calc(100% - 10px)",
+                            margin="5px",
+                        ),
+                        debounce=True,
+                    )
                 ),
                 Pane(
                     list(sidebar_components.values()),
@@ -411,7 +467,7 @@ class BaseStudio:
                     ],
                     value="result",
                     persistence=True,
-                    inputStyle=dict(marginLeft="1rem"),
+                    inputStyle=dict(marginLeft="10px", marginRight="2px"),
                 ),
             ],
             style=dict(
@@ -507,7 +563,7 @@ class BaseStudio:
         result = results[function_name]
 
         error = None
-        if result_processor_value:
+        if result_processor_value.strip():
             try:
                 result = self.process_result(result, result_processor_value)
             except Exception as e:
@@ -535,8 +591,19 @@ class BaseStudio:
         )
 
     def populate_definition(self, composer, function_name):
+
         source = composer.get_source(function_name)
-        return function_name, None, None, html.Pre(source, style=dict(padding="0.5rem"))
+        highlighted = highlight(source, PythonLexer(), HtmlFormatter())
+
+        return (
+            function_name,
+            None,
+            None,
+            html.Div(
+                dash_dangerously_set_inner_html.DangerouslySetInnerHTML(highlighted),
+                style=dict(padding="0.5rem"),
+            ),
+        )
 
     def populate_profiler(self, composer, function_name, parameters):
 
@@ -571,9 +638,9 @@ class BaseStudio:
                         background=color,
                         height="1rem",
                     ),
-                    title=f"{key} = {metric_total:.2f}%",
+                    title=f"{key} = {metric_total:.2f}% ({metric_absolute:.3f}s)",
                 )
-                for key, value, metric_total, color in metrics
+                for key, value, metric_total, metric_absolute, color in metrics
             ]
 
         def profile_section(section, bar_description):
@@ -588,6 +655,7 @@ class BaseStudio:
                                         key,
                                         metrics[key] / highest * 90,
                                         metrics[key] / total * 100,
+                                        metrics[key],
                                         color,
                                     )
                                     for key, color in bar_description
@@ -652,17 +720,15 @@ class BaseStudio:
             return self.populate_definition(composer, function_name)
         else:
             return self.populate_profiler(
-                composer,
-                function_name,
-                dict(zip(parameter_keys, parameter_values)),
+                composer, function_name, dict(zip(parameter_keys, parameter_values))
             )
 
     def update_composer_parameters(self, composer, parameters):
         def smartish_cast(type_, value):
             if issubclass(type_, bool) and isinstance(value, str):
-                return value.lower() == "t"
-
-            return value
+                return value.lower()[0] == "t"
+            else:
+                return value
 
         parameters = {
             key: smartish_cast(type_, parameters[key])
@@ -674,6 +740,7 @@ class BaseStudio:
     def populate_graph(
         self,
         composer,
+        node_name_filter,
         graph_display_options,
         graph_neighbourhood,
         graph_neighbourhood_size,
@@ -729,6 +796,13 @@ class BaseStudio:
                     )
                 )
 
+        if node_name_filter:
+            subgraph = {
+                node
+                for node in subgraph
+                if node_name_filter.strip().lower() in node.lower()
+            }
+
         if caching:
             instructions = get_execution_instructions(composer, composer.dag(), [])
 
@@ -768,11 +842,19 @@ class ExternalStudio(BaseStudio):
     """
 
     def result_processor(self):
-        return dcc.Textarea(
+        return dash_ace_persistent.DashAceEditor(
             id="result-processor",
+            value="",
+            theme="github",
+            mode="python",
+            tabSize=2,
             placeholder='Enter a query string.\n\nYou can use full pandas query strings.\ne.g.: merchant_id == "ABC"',
-            rows=5,
-            style=dict(width="100%", border="none"),
+            maxLines=10,
+            minLines=5,
+            showGutter=False,
+            persistence=True,
+            highlightActiveLine=False,
+            debounceChangePeriod=1000,
         )
 
     def process_result(self, result, value):
